use crate::{locale::LocaleType, Pool};
use sqlx::{prelude::FromRow, query, query_as, Executor, MySql};

#[derive(Clone, Debug, PartialEq, FromRow, Eq)]
pub struct MasterPass {
    pub user_id: u64,
    pub salt: Vec<u8>,
    pub password_hash: Vec<u8>,
}

impl MasterPass {
    // Inserts the master password into DB
    #[inline]
    pub async fn insert(&self, pool: &Pool, locale: LocaleType) -> crate::Result<()> {
        let locale: u8 = locale.into();
        query!(
            "INSERT INTO master_pass VALUES (?, ?, ?, ?)",
            self.user_id,
            self.salt,
            self.password_hash,
            locale
        )
        .execute(pool)
        .await
        .map(|_| ())
    }

    /// Gets the master password from the database
    #[inline]
    pub async fn get(user_id: u64, pool: &Pool) -> crate::Result<Option<Self>> {
        query_as("SELECT user_id, salt, password_hash FROM master_pass WHERE user_id = ?")
            .bind(user_id)
            .fetch_optional(pool)
            .await
    }

    /// Checks if the master password for the user exists
    #[inline]
    pub async fn exists(user_id: u64, pool: &Pool) -> crate::Result<bool> {
        query_as::<_, (bool,)>(
            "SELECT EXISTS(SELECT * FROM master_pass WHERE user_id = ? LIMIT 1) as value",
        )
        .bind(user_id)
        .fetch_one(pool)
        .await
        .map(|(exists,)| exists)
    }

    /// Removes a master password of the user from the database
    pub async fn remove(
        user_id: u64,
        pool: impl Executor<'_, Database = MySql>,
    ) -> crate::Result<()> {
        query!("DELETE FROM master_pass WHERE user_id = ?", user_id)
            .execute(pool)
            .await
            .map(|_| ())
    }
}
